'''
Contains a class that represents a server-side peer, i.e. a connection to a client.
'''
class ClientConnection:
    '''
    A connection to a Wireguard client. Contains the public key of the client,
    and the local IP of the client.
    '''
    def __init__(self, public_key, local_ip):
        self.public_key = public_key
        self.local_ip = local_ip

    def get_key(self):
        '''
        Get the public key of the client this points to.
        '''
        return self.public_key

    def get_ip(self):
        '''
        Get the local IP.
        '''
        return self.local_ip



'''
Contains a class that represents a client-side peer, i.e. a connection to a server.
'''
class ServerConnection:
    '''
    A connection to a Wireguard server. Contains the public key of the server,
    and the endpoint IP and port.
    '''
    def __init__(self, public_key, host_ip, port):
        self.public_key = public_key
        self.host_ip = host_ip
        self.port = port

    def get_key(self):
        '''
        Get the public key of the server this points to.
        '''
        return self.public_key

    def get_endpoint(self):
        '''
        Get the destination IP.
        '''
        return self.host_ip

    def get_port(self):
        '''
        Get the server port.
        '''
        return self.port



'''
Contains a class representing a client-side Wireguard connection.
'''
from .wireguard import valid_interface, create_client, delete_device, client_add_peer, enable_device
from .key import Key
from .server_connection import ServerConnection

class Client:
    '''
    This is a client-side Wireguard connection.
    '''
    def __init__(self, interface_name, key, local_ip):
        if not valid_interface(interface_name):
            raise ValueError(f"Invalid interface name {interface_name}")
        if not isinstance(key, Key):
            raise ValueError("Key should be an instance of python_wireguard.Key")

        self.key = key
        self.local_ip = local_ip
        self.interface_name = interface_name
        self.connection = None
        self.interface_created = False

    def create_interface(self):
        '''
        Create the internet interface belonging to this client.
        '''
        create_client(self.interface_name, self.key.as_bytes(), self.local_ip)
        self.interface_created = True

    def delete_interface(self):
        '''
        Delete the internet interface belonging to this client.
        '''
        delete_device(self.interface_name)
        self.interface_created = False

    def set_server(self, server_connection):
        '''
        Set the server for this Client.
        '''
        if not isinstance(server_connection, ServerConnection):
            raise ValueError(
                "server_connection should be an instance of python_wireguard.ServerConnection"
                )
        self.connection = server_connection

    def connect(self):
        '''
        Connect to the Wireguard server.
        '''
        if not self.interface_created:
            self.create_interface()
        if self.connection is None:
            raise ValueError(
                "The connection has not been configured for this Client yet. \
                Use 'set_server' to set it."
                )
        server_connection = self.connection
        key = server_connection.get_key()
        client_add_peer(self.interface_name, key.as_bytes(),
            server_connection.get_endpoint(), server_connection.get_port())
        enable_device(self.interface_name)



'''
This file contains functions for directly interacting with the Wireguard shared objects file.
'''
from ctypes import CDLL, c_ushort, create_string_buffer
import re
import os

dirname = os.path.dirname(os.path.realpath(__file__))
c_library = CDLL(dirname + '/bin/py-wireguard.so')

def valid_interface(interface):
    '''
    Validate the name of a network interface.
    '''
    return bool(re.match("^[A-Za-z0-9_-]*$", interface))

def empty_key():
    '''
    Create an empty key. Used for generating new keys, you should not need this.
    '''
    return create_string_buffer(b'\000' * 32)

def key_pair():
    '''
    Create a private/public key pair fo ruse with Wireguard.
    '''
    private, public = empty_key(), empty_key()
    c_library.generate_private_key(private)
    c_library.generate_public_key(private, public)
    return private, public

def create_server(name, port, private_key, local_ip):
    '''
    Create a server-side Wireguard interface. This is a Wireguard instance that listens on a port
    to allow clients to connect.
    '''
    if valid_interface(name):
        c_library.add_server_device(name.encode(), c_ushort(port), private_key)
        os.system(f"ip a add dev {name} {local_ip}")
    else:
        print(f"invalid device name '{name}'")

def create_client(name, private_key, local_ip):
    '''
    Create a client-side Wireguard interface. This means that it won't be listening on a port.
    '''
    if valid_interface(name):
        c_library.add_client_device(name.encode(), private_key)
        os.system(f"ip a add dev {name} {local_ip}")
    else:
        print(f"invalid device name '{name}'")

def client_add_peer(device_name, public_key, address, port):
    '''
    Add a 'peer' to a client-side setup. This means that the server will be added for this setup.
    '''
    c_library.add_server_peer(device_name.encode(), public_key, address.encode(), c_ushort(port))

def server_add_peer(device_name, public_key, local_ip):
    '''
    Add a client
    '''
    c_library.add_client_peer(device_name.encode(), public_key, local_ip.encode())

def delete_device(name):
    '''
    Delete interface :name:
    '''
    c_library.delete_device(name.encode())

def list_devices():
    '''
    Print a list of all Wireguard network devices.
    '''
    c_library.list_devices()

def key_to_base64(key):
    '''
    Convert a binary key to base64.
    '''
    str_val = create_string_buffer(b'\000' * 44)
    c_library.key_to_string(key, str_val)
    return str_val.value.decode()

def key_from_base64(base64):
    '''
    Create a binary key from a base64 string.
    '''
    key = empty_key()
    c_library.key_from_string(base64.encode(), key)
    return key

def enable_device(device):
    '''
    Turn on a network device with name :device:.
    '''
    if valid_interface(device):
        os.system(f"ip link set up {device}")
    else:
        print(f"invalid device '{device}'")



"""
Contains a class representing a server-side Wireguard connection.
"""
from .wireguard import valid_interface, create_server, delete_device, server_add_peer, enable_device
from .key import Key
from .client_connection import ClientConnection


class Server:
    """
    This is a server-side Wireguard connection.
    """

    def __init__(self, interface_name, key, local_ip, port):
        if not valid_interface(interface_name):
            raise ValueError(f"Invalid interface name {interface_name}")
        if not isinstance(key, Key):
            raise ValueError("Key should be an instance of python_wireguard.Key")
        self.interface_name = interface_name
        self.key = key
        self.local_ip = local_ip
        self.port = port
        self.clients = []
        self.interface_created = False

    def create_interface(self):
        """
        Create the network interface belonging to this wg server.
        """
        create_server(self.interface_name, self.port, self.key.as_bytes(), self.local_ip)
        self.interface_created = True

    def delete_interface(self):
        """
        Deletes the network interface of this server.
        """
        delete_device(self.interface_name)
        self.interface_created = False

    def add_client(self, client_connection):
        """
        Add a new client to this server.
        """
        if not isinstance(client_connection, ClientConnection):
            raise ValueError(
                "server_connection should be an instance of python_wireguard.ServerConnection"
            )
        server_add_peer(self.interface_name,
                        client_connection.get_key().as_bytes(),
                        client_connection.get_ip())

    def enable(self):
        """
        Turn on the network interface.
        """
        if not self.interface_created:
            self.create_interface()
        enable_device(self.interface_name)



'''
Python class for using a Wireguard public or private key.
'''
import python_wireguard.wireguard as wg

class Key:
    '''
    This represents a Wireguard key.
    '''
    def __init__(self, base64_string=None):
        if base64_string is None:
            self.byte_content = wg.empty_key()
        elif len(base64_string) != 44:
            raise ValueError("Invalid base64 string provided")
        else:
            self.byte_content = wg.key_from_base64(base64_string)

    def __str__(self):
        '''
        Convert this key to a string.
        '''
        return wg.key_to_base64(self.byte_content)

    def __repr__(self):
        '''
        Represent this class (used in the python cli).
        '''
        return f"<python_wireguard.Key object: {str(self)}>"

    @staticmethod
    def key_pair():
        '''
        Create a new private/public key pair.
        '''
        private, public = wg.key_pair()
        private_obj = Key(wg.key_to_base64(private))
        public_obj = Key(wg.key_to_base64(public))
        return private_obj, public_obj

    def as_bytes(self):
        '''
        Get the c_char_array content of this key.
        '''
        return self.byte_content



'''
This module is a wrapper for interacting with Wireguard on Linux.
'''
from .wireguard import list_devices, delete_device
from .key import Key
from .client import Client
from .server_connection import ServerConnection
from .server import Server
from .client_connection import ClientConnection



'''
Used for building the python_wireguard package.
'''
from pathlib import Path
import os
from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

version = os.environ.get('RELEASE_VERSION')

setup(name='python_wireguard',
      version=version,
      description='A python wrapper for controlling Wireguard',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/jarnoaxel/python-wireguard',
      author='jarnoaxel',
      license='GPL-3.0-or-later',
      packages=['python_wireguard'],
      include_package_data=True,
      keywords=['vpn', 'wireguard'],
      zip_safe=False)
